//*********************************************** Gametrak stuff ***********************************************/

// z axis deadzone
.032 => float DEADZONE;

// which joystick
0 => int device;
// get from command line
if( me.args() ) me.arg(0) => Std.atoi => device;

// data structure for gametrak
class GameTrak
{
    // timestamps
    time lastTime;
    time currTime;
    
    //2 messages ago axis data
    float lastLastAxis[6];
    // previous axis data
    float lastAxis[6];
    // current axis data
    float axis[6];
}
// gametrack
GameTrak gt;
int isBtnDown;

// HID objects
Hid trak;
HidMsg msg;

// open joystick 0, exit on fail
if( !trak.openJoystick( device ) ) me.exit();

// print
<<< "joystick '" + trak.name() + "' ready", "" >>>;

// spork control
spork ~ gametrak();
// print
//spork ~ print();

// print
fun void print()
{
    // time loop
    while( true )
    {
        // values
        <<< "axes:", gt.axis[0],gt.axis[1],gt.axis[2], gt.axis[3],gt.axis[4],gt.axis[5] >>>;
        // advance time
        100::ms => now;
    }
}

// gametrack handling
fun void gametrak()
{
    while( true )
    {
        // wait on HidIn as event
        trak => now;
        
        // messages received
        while( trak.recv( msg ) )
        {
            // joystick axis motion
            if( msg.isAxisMotion() )
            {            
                // check which
                if( msg.which >= 0 && msg.which < 6 )
                {
                    // check if fresh
                    if( now > gt.currTime )
                    {
                        // time stamp
                        gt.currTime => gt.lastTime;
                        // set
                        now => gt.currTime;
                    }
                    // save last-last
                    gt.lastAxis[msg.which] => gt.lastLastAxis[msg.which];
                    // save last
                    gt.axis[msg.which] => gt.lastAxis[msg.which];
                    // the z axes map to [0,1], others map to [-1,1]
                    if( msg.which != 2 && msg.which != 5 )
                    { msg.axisPosition => gt.axis[msg.which]; }
                    else
                    {
                        1 - ((msg.axisPosition + 1) / 2) - DEADZONE => gt.axis[msg.which];
                        if( gt.axis[msg.which] < 0 ) 0 => gt.axis[msg.which];
                    }
                }
            }
            
            // joystick button down
            else if( msg.isButtonDown() )
            {
                1 => isBtnDown;
                <<< "button", msg.which, "down" >>>;
            }
            
            // joystick button up
            else if( msg.isButtonUp() )
            {
                0 => isBtnDown;
                <<< "button", msg.which, "up" >>>;
            }
        }
    }
}



//*********************************************** End Gametrak stuff ***********************************************/

SqrOsc s => dac;
0.3 => s.gain;
float freqs[];
[220.0, 440, 660, 880] @=> freqs;
0 => int curIndex;
0.1::second => dur cooldownTime;
now => time lastChange;
0.0004 => float requiredDiff;

while(true)
{
    freqs[curIndex] => s.freq;
    
    gt.axis[5] - gt.lastAxis[5] => float downDiff;
    gt.lastLastAxis[5] - gt.lastAxis[5] => float upDiff;
    <<<downDiff, upDiff>>>;
    if(now - lastChange > cooldownTime && downDiff > requiredDiff && upDiff > 0)
    {
        curIndex++;
        if(curIndex > 3)
            0 => curIndex;
        now => lastChange;
    }
    10::ms => now;
}